package com.trifork.sosigw.sts.servlet;

import com.trifork.sosigw.Factory;
import com.trifork.sosigw.api.IDCardCache;
import com.trifork.sosigw.api.IDCardState;
import com.trifork.sosigw.cluster.IDCardClusterCacheEntry;
import com.trifork.sosigw.config.RuntimeConfig;
import com.trifork.sosigw.sts.STSClusterProxy;
import com.trifork.sosigw.sts.STSProxy;
import com.trifork.sosigw.util.Util;
import dk.sosi.seal.model.OIOBSTSAMLAssertionToIDCardRequest;
import dk.sosi.seal.model.OIOBSTSAMLAssertionToIDCardResponse;
import dk.sosi.seal.model.UserIDCard;
import dk.sosi.seal.model.constants.DSTags;
import dk.sosi.seal.model.constants.SOAPTags;
import dk.sosi.seal.model.constants.SOSIAttributes;
import dk.sosi.seal.modelbuilders.OIOBSTSAMLAssertionToIDCardRequestModelBuilder;
import dk.sosi.seal.modelbuilders.OIOBSTSAMLAssertionToIDCardResponseModelBuilder;
import dk.sosi.seal.xml.XmlUtil;
import org.apache.log4j.Logger;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;
import org.xml.sax.InputSource;

import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.io.IOException;
import java.util.Properties;

public class CreateIdCardFromBSTRequestHandler {
    public static final Logger log = Logger.getLogger(CreateIdCardFromBSTRequestHandler.class);
    // We should only create the sts proxies once, so make stsService static
    protected static STSProxy stsService =
            STSClusterProxy.create(RuntimeConfig.getSecurityTokenServiceUrlList());

    private final HttpServletRequest req;
    private final HttpServletResponse resp;

    public CreateIdCardFromBSTRequestHandler(HttpServletRequest req, HttpServletResponse resp) {
        this.req = req;
        this.resp = resp;
    }

    public void handleRequest() throws IOException, ServletException, InvalidRequestException {
        log.debug("createIdCardFromBST");
        resp.setContentType("text/xml; charset=utf-8");
        Document requestDocument = XmlUtil.readXml(new Properties(), new InputSource(req.getInputStream()), false);
        OIOBSTSAMLAssertionToIDCardRequest request = buildAndValidateOIOBSTSAMLAssertionToIDCardRequest(requestDocument);
        String responseString = stsService.callBST2SOSI(requestDocument);
        Document responseDocument = XmlUtil.readXml(new Properties(), responseString, false);
        OIOBSTSAMLAssertionToIDCardResponse response = new OIOBSTSAMLAssertionToIDCardResponseModelBuilder().build(responseDocument);
        if (response.isFault()) {
            log.info("bst2sosi request failed with error message: " + response.getFaultString());
        } else {
            UserIDCard idCard = getUserIDCardFromOIOBSTSAMLAssertionToIDCardResponse(response);
            String cacheKey = getCacheKeyFromRequest(request);
            saveIdCardInCache(idCard, cacheKey);
            removeSignaturesFromBst2SosiResponse(responseDocument);
        }

        byte[] responseBytes = XmlUtil.serializeXml2ByteArray(responseDocument, false);
        resp.getOutputStream().write(responseBytes);
        resp.setStatus(200);
    }

    private static void removeSignaturesFromBst2SosiResponse(Document responseDocument) throws ServletException {
        Element responseDocumentElement = responseDocument.getDocumentElement();
        Element header = (Element) responseDocumentElement.getElementsByTagName(SOAPTags.HEADER_PREFIXED).item(0);
        NodeList signatureNodes = header.getElementsByTagName(DSTags.SIGNATURE_PREFIXED);
        if (signatureNodes.getLength() != 1) {
            throw new ServletException("bst2sosi response header should contain 1 signature element, but contained " + signatureNodes.getLength());
        }
        Element signature = (Element) signatureNodes.item(0);
        signature.getParentNode().removeChild(signature);

        Element body = (Element) responseDocumentElement.getElementsByTagName(SOAPTags.BODY_PREFIXED).item(0);
        signatureNodes = body.getElementsByTagName(DSTags.SIGNATURE_PREFIXED);
        if (signatureNodes.getLength() != 1) {
            throw new ServletException("bst2sosi response body should contain 1 signature element, but contained " + signatureNodes.getLength());
        }
        signature = (Element) signatureNodes.item(0);
        signature.getParentNode().removeChild(signature);
    }

    private static void saveIdCardInCache(UserIDCard idCard, String cacheKey) {
        final IDCardCache idCardCache = Factory.getIDCardCache();
        final IDCardClusterCacheEntry cachedIDCardEntry = idCardCache.ensureIDCard(cacheKey);
        cachedIDCardEntry.setIdCard(idCard);
        idCardCache.updateEntryState(cachedIDCardEntry, IDCardState.SIGNED);
        idCardCache.entryUpdatedLocally(cachedIDCardEntry);
    }

    private static String getCacheKeyFromRequest(OIOBSTSAMLAssertionToIDCardRequest request) {
        String nameId = request.getClaimMap().get(SOSIAttributes.SUBJECT_NAME_ID);
        nameId = nameId == null ? request.getOIOBSTSAMLAssertion().getSubjectNameID() : nameId;
        return Util.computeCacheKey(nameId);
    }

    private static UserIDCard getUserIDCardFromOIOBSTSAMLAssertionToIDCardResponse(OIOBSTSAMLAssertionToIDCardResponse response) throws InvalidRequestException {
        if (!(response.getIDCard() instanceof UserIDCard)) {
            throw new InvalidRequestException("createIdCardFromBST only supports user id cards");
        }
        return (UserIDCard) response.getIDCard();
    }

    private static OIOBSTSAMLAssertionToIDCardRequest buildAndValidateOIOBSTSAMLAssertionToIDCardRequest(Document requestDoc) throws InvalidRequestException {
        OIOBSTSAMLAssertionToIDCardRequest request = new OIOBSTSAMLAssertionToIDCardRequestModelBuilder().build(requestDoc);
        if (request.isAssertionEncrypted() && !request.getClaimMap().containsKey(SOSIAttributes.SUBJECT_NAME_ID)) {
            throw new InvalidRequestException("createIdCardFromBST with encrypted bootstraptoken must include a SubjectNameID claim");
        }
        return request;
    }
}
