/*
* The MIT License
*
* Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Permission is hereby granted, free of charge, to any person obtaining a copy of
* this software and associated documentation files (the "Software"), to deal in
* the Software without restriction, including without limitation the rights to
* use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
* of the Software, and to permit persons to whom the Software is furnished to do
* so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in all
* copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
* OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
* SOFTWARE.
*
* $HeadURL$
* $Id$
*/
package com.trifork.sosigw.ws;

import java.math.BigDecimal;
import java.util.Calendar;
import java.util.GregorianCalendar;

import oasis.names.tc.saml._2_0.assertion.AssertionType;
import oasis.names.tc.saml._2_0.assertion.Attribute;
import oasis.names.tc.saml._2_0.assertion.AttributeStatement;
import oasis.names.tc.saml._2_0.assertion.Conditions;
import oasis.names.tc.saml._2_0.assertion.NameIDType;
import oasis.names.tc.saml._2_0.assertion.Subject;
import oasis.names.tc.saml._2_0.assertion.SubjectConfirmation;
import oasis.names.tc.saml._2_0.assertion.SubjectConfirmationData;

import org.w3._2000._09.xmldsig_.KeyInfo;

import com.sun.org.apache.xerces.internal.jaxp.datatype.XMLGregorianCalendarImpl;
import com.trifork.sosigw.config.StaticConfigApi;
import com.trifork.sosigw.config.StaticConfigImpl;

import dk.sosi.seal.model.CareProvider;
import dk.sosi.seal.model.IDCard;
import dk.sosi.seal.model.SystemInfo;
import dk.sosi.seal.model.UserIDCard;
import dk.sosi.seal.model.UserInfo;
import dk.sosi.seal.model.constants.IDValues;
import dk.sosi.seal.model.constants.MedcomAttributes;
import dk.sosi.seal.model.constants.SOSIAttributes;
import dk.sosi.seal.model.constants.SubjectIdentifierTypeValues;

/**
 *
 *
 * @author Jeppe Sommer <jso at trifork.com>
 * @author Morten Grouleff <mgr at trifork.com>
 */
public class IDCardToSamlAssertion {
    private static final StaticConfigApi STATIC_CONF = StaticConfigImpl.getInstance();

    //    static final boolean USE_LOCALTIME_IN_SAML = true; // DGWS-1.0 says localtime in all timestamps. Bad stuff.

    private static final String HOLDER_OF_KEY = "urn:oasis:names:tc:SAML:2.0:cm:holder-of-key";


    public static Attribute makeAttribute(String name, String format, String value) {
        Attribute attribute = new Attribute();
        attribute.setName(name);
        if (format != null) {
            attribute.setNameFormat(format);
        }
        attribute.setAttributeValue(value);
        return attribute;
    }

    /**
     * Make a saml:assertion out of an idcard, suitable for sending as part of a webservice soap header.
     * Modeled closely on IDCardDOMBuilder.java in SOSI-SEAL, which makes a DOM of an idcard.
     * @param idCard
     * @return
     */
    public static AssertionType build(UserIDCard idCard) {
        final UserInfo userInfo = idCard.getUserInfo();
        final SystemInfo systemInfo = idCard.getSystemInfo();

        AssertionType a = new AssertionType();

        a.setVersion(new BigDecimal("2.0"));
        a.setId(IDValues.IDCARD);
        a.setIssuer(idCard.getIssuer());

        Subject subject = new Subject();
        NameIDType nameID = new NameIDType();
        nameID.setFormat(SubjectIdentifierTypeValues.CPR_NUMBER);
        nameID.setValue(userInfo.getCPR());
        subject.setNameID(nameID);

        SubjectConfirmation subjectConfirmation = new SubjectConfirmation();

        subjectConfirmation.setConfirmationMethod(HOLDER_OF_KEY);
        SubjectConfirmationData subjectConfirmationData = new SubjectConfirmationData();
        KeyInfo keyInfo = new KeyInfo();
        keyInfo.setKeyName(IDValues.OCES_SIGNATURE);
        subjectConfirmationData.setKeyInfo(keyInfo);
        subjectConfirmation.setSubjectConfirmationData(subjectConfirmationData);
        subject.setSubjectConfirmation(subjectConfirmation);
        a.setSubject(subject);


        AttributeStatement stmt;

        stmt = new AttributeStatement();
        stmt.setId(IDValues.IDCARD_DATA);
        addNullableAttribute(stmt, SOSIAttributes.IDCARD_ID, idCard.getIDCardID());
        addNullableAttribute(stmt, SOSIAttributes.IDCARD_VERSION, idCard.getVersion());
        addNullableAttribute(stmt, SOSIAttributes.IDCARD_TYPE, IDCard.IDCARDTYPE_USER);
        addNullableAttribute(stmt, SOSIAttributes.AUTHENTICATION_LEVEL, "" + idCard.getAuthenticationLevel().getLevel());
        addNullableAttribute(stmt, SOSIAttributes.OCES_CERT_HASH, idCard.getCertHash());
        a.getAttributeStatement().add(stmt);

        stmt = new AttributeStatement();
        stmt.setId(IDValues.USER_LOG);
        if (!"".equals(userInfo.getCPR())) {
            addNullableAttribute(stmt, MedcomAttributes.USER_CIVIL_REGISTRATION_NUMBER, userInfo.getCPR());
        }
        addNullableAttribute(stmt, MedcomAttributes.USER_GIVEN_NAME, userInfo.getGivenName());
        addNullableAttribute(stmt, MedcomAttributes.USER_SURNAME, userInfo.getSurName());
        addNullableAttribute(stmt, MedcomAttributes.USER_EMAIL_ADDRESS, userInfo.getEmail());
        addNullableAttribute(stmt, MedcomAttributes.USER_ROLE, userInfo.getRole());
        addNullableAttribute(stmt, MedcomAttributes.USER_OCCUPATION, userInfo.getOccupation());
        addNullableAttribute(stmt, MedcomAttributes.USER_AUTHORIZATION_CODE, userInfo.getAuthorizationCode());
        a.getAttributeStatement().add(stmt);

        stmt = new AttributeStatement();
        stmt.setId(IDValues.SYSTEM_LOG);
        addNullableAttribute(stmt, MedcomAttributes.IT_SYSTEM_NAME, systemInfo.getITSystemName());
        CareProvider careProvider = systemInfo.getCareProvider();
        stmt.getAttribute().add(makeAttribute(MedcomAttributes.CARE_PROVIDER_ID, careProvider.getType(), careProvider.getID()));
        addNullableAttribute(stmt, MedcomAttributes.CARE_PROVIDER_NAME, careProvider.getOrgName());
        a.getAttributeStatement().add(stmt);

        Conditions conditions = new Conditions();
        Calendar notBefore = Calendar.getInstance();
        notBefore.setTime(idCard.getCreatedDate());
        if (STATIC_CONF.useLocalTimeInSaml()) {
            notBefore.add(Calendar.MILLISECOND, notBefore.getTimeZone().getOffset(notBefore.getTimeInMillis()));
        }

        XMLGregorianCalendarImpl notBeforeGregorianCalendarImpl = new XMLGregorianCalendarImpl((GregorianCalendar)notBefore);
        conditions.setNotBefore(notBeforeGregorianCalendarImpl);
        a.setIssueInstant(notBeforeGregorianCalendarImpl);

        Calendar notOnOrAfter = Calendar.getInstance();
        notOnOrAfter.setTime(idCard.getExpiryDate());
        if (STATIC_CONF.useLocalTimeInSaml()) {
            notOnOrAfter.add(Calendar.MILLISECOND, notOnOrAfter.getTimeZone().getOffset(notOnOrAfter.getTimeInMillis()));
        }
        conditions.setNotOnOrAfter(new XMLGregorianCalendarImpl((GregorianCalendar)notOnOrAfter));
        a.setConditions(conditions);

        return a;
    }

    private static void addNullableAttribute(AttributeStatement stmt, String attribute, String attributeValue) {
        if (attributeValue != null) {
            stmt.getAttribute().add(makeAttribute(attribute, null, attributeValue));
        }
    }
}