package com.trifork.sosigw.inttest;

import com.trifork.sosigw.config.StaticConfigImpl;
import com.trifork.sosigw.nts.NTSClient;
import com.trifork.sosigw.ws.IDCardToSamlAssertion;
import com.trifork.sosigw.ws.IdCardTestUtils;
import dk.medcom.dgws._2006._04.dgws_1_0.Header;
import dk.nspop.nts._2013._05.NTS;
import dk.sosi.seal.model.OIOBSTSAMLAssertionToIDCardResponse;
import dk.sosi.seal.model.UserIDCard;
import dk.sosi.seal.model.constants.DSTags;
import dk.sosi.seal.model.constants.SubjectIdentifierTypeValues;
import dk.sosi.seal.modelbuilders.OIOBSTSAMLAssertionToIDCardResponseModelBuilder;
import dk.sosi.seal.xml.XmlUtil;
import oasis.names.tc.saml._2_0.assertion.AssertionType;
import org.junit.After;
import org.junit.Assert;
import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.Test;
import org.oasis_open.docs.wss._2004._01.oasis_200401_wss_wssecurity_secext_1_0.Security;
import org.w3c.dom.Document;
import org.w3c.dom.Node;

import java.io.IOException;
import java.net.MalformedURLException;
import java.util.Properties;
import java.util.Scanner;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;
import static org.junit.Assert.assertTrue;

public class CreateIdCardFromBstIT extends AbstractIT {
    private static final String NTS_SERVER_URL = System.getProperty("ntsUrl", "https://test1.ekstern-test.nspop.dk:8443");
    private static final String SOSIGW_SERVER_URL = System.getProperty("sosigwUrl", "http://localhost:8080");
    private static final String REQUEST_URL = SOSIGW_SERVER_URL + "/sosigw/service/sosigw";
    private static final String CREATE_IDCARD_FROM_BST_URL = REQUEST_URL + "/createIdCardFromBST";
    private static NTS nts;
    private static Header medcomHeader;

    @BeforeClass
    public static void setupNTS() throws MalformedURLException {
        StaticConfigImpl.initSettings();
        NTSClient ntsClient = new NTSClient();
        ntsClient.configure(new String[]{"-g", SOSIGW_SERVER_URL, "-h", NTS_SERVER_URL}, new Properties());
        nts = ntsClient.getNTSPort();
        medcomHeader = ntsClient.createMedcomHeader();
    }

    @Before
    @After
    public void clearCache() throws Exception {
        String logoutRequestBody = getResourceString("logout-Mads-Skjern-request.xml");
        sendRequest(logoutRequestBody, REQUEST_URL, false);
        logoutRequestBody = getResourceString("logout-Korsbaek-request.xml");
        sendRequest(logoutRequestBody, REQUEST_URL, false);
    }

    /**
     * GW-T1
     * Prækondition: Bruger har ikke et idkort i GW-cachen i forvejen
     * Input: Validt createIdCardFromBST request uden angivelse af et ”sosi:SubjectNameID” claim
     * Output: Validt svar med det nyudstedte idkort i usigneret form
     * Sideeffekt: Det nyudstedte idkort er gemt i cachen under SubjectNameID fra bootstraptokenet og kan benyttes til at kalde en test-service (fx NTS) via GW
     */
    @Test
    public void testValidRequestWithoutNameIdClaimEmptyCache() throws Exception  {
        // GIVEN
        String requestBody = getResourceString("../sts/servlet/OIOH3BST2SOSI-without-nameid-claim-request.xml");

        // WHEN
        Document responseDoc = sendRequest(requestBody, CREATE_IDCARD_FROM_BST_URL, true);

        // THEN
        validateResponse(responseDoc, "KorsbaekKommune\\MSK");
        String getValidIdCardRequestBody = getResourceString("getValidIdCard-Korsbaek-request.xml");
        sendRequest(getValidIdCardRequestBody, REQUEST_URL, true); // Fails if idcard is not in cache
    }

    /**
     * GW-T2
     * Prækondition: Bruger har ikke et idkort i GW-cachen i forvejen
     * Input: Validt createIdCardFromBST request med angivelse af et ”sosi:SubjectNameID” claim (som er forskelligt fra SubjectNameID i BST)
     * Output: Validt svar med det nyudstedte idkort i usigneret form
     * Sideeffekt: Det nyudstedte idkort er gemt i cachen under SubjectNameID fra claimet og kan benyttes til at kalde en test-service (fx NTS) via GW
     */
    @Test
    public void testValidRequestWithNameIdClaimEmptyCache() throws Exception  {
        // GIVEN
        String requestBody = getResourceString("../sts/servlet/OIOH3BST2SOSI-with-nameid-claim-request.xml");

        // WHEN
        Document responseDoc = sendRequest(requestBody, CREATE_IDCARD_FROM_BST_URL, true);

        // THEN
        validateResponse(responseDoc, "Mads_Skjern");
        String getValidIdCardRequestBody = getResourceString("getValidIdCard-Mads-Skjern-request.xml");
        sendRequest(getValidIdCardRequestBody, REQUEST_URL, true); // Fails if idcard is not in cache
    }

    /**
     * GW-T3
     * Prækondition: Bruger har et idkort i GW-cachen i forvejen
     * Input: Validt createIdCardFromBST request uden angivelse af et ”sosi:SubjectNameID” claim
     * Output: Validt svar med det nyudstedte idkort i usigneret form
     * Sideeffekt: Det nyudstedte idkort er gemt i cachen under SubjectNameID fra bootstraptokenet, har erstattet det eksisterende idkort og kan benyttes til at kalde en test-service (fx NTS) via GW
     */
    @Test
    public void testValidRequestWithoutNameIdClaimWithIdCardInCache() throws Exception  {
        // GIVEN
        String requestBody = getResourceString("../sts/servlet/OIOH3BST2SOSI-without-nameid-claim-request.xml");
        Document responseDoc = sendRequest(requestBody, CREATE_IDCARD_FROM_BST_URL, true);
        OIOBSTSAMLAssertionToIDCardResponse response = parseResponse(responseDoc);
        assertFalse(response.getFaultString(), response.isFault());

        // WHEN
        responseDoc = sendRequest(requestBody, CREATE_IDCARD_FROM_BST_URL, true);

        // THEN
        validateResponse(responseDoc, "KorsbaekKommune\\MSK");
        String getValidIdCardRequestBody = getResourceString("getValidIdCard-Korsbaek-request.xml");
        sendRequest(getValidIdCardRequestBody, REQUEST_URL, true); // Fails if idcard is not in cache
    }

    /**
     * GW-T4
     * Prækondition: Bruger har et idkort i GW-cachen i forvejen
     * Input: Validt createIdCardFromBST request med angivelse af et ”sosi:SubjectNameID” claim (som er forskelligt fra SubjectNameID i BST)
     * Output: Validt svar med det nyudstedte idkort i usigneret form
     * Sideeffekt: Det nyudstedte idkort er gemt i cachen under SubjectNameID fra claimet, har erstattet det eksisterende idkort og kan benyttes til at kalde en test-service (fx NTS) via GW
     */
    @Test
    public void testValidRequestWithNameIdClaimWithIdCardInCache() throws Exception  {
        // GIVEN
        String requestBody = getResourceString("../sts/servlet/OIOH3BST2SOSI-with-nameid-claim-request.xml");
        Document responseDoc = sendRequest(requestBody, CREATE_IDCARD_FROM_BST_URL, true);
        OIOBSTSAMLAssertionToIDCardResponse response = parseResponse(responseDoc);
        assertFalse(response.getFaultString(), response.isFault());

        // WHEN
        responseDoc = sendRequest(requestBody, CREATE_IDCARD_FROM_BST_URL, true);

        // THEN
        validateResponse(responseDoc, "Mads_Skjern");
        String getValidIdCardRequestBody = getResourceString("getValidIdCard-Mads-Skjern-request.xml");
        sendRequest(getValidIdCardRequestBody, REQUEST_URL, true); // Fails if idcard is not in cache
    }

    /**
     * GW-T5
     * Prækondition: Bruger har et idkort i GW-cachen i forvejen
     * Input: Invalidt createIdCardFromBST request
     * Output: Fejlsvar uden idkort (videresendt i uforandret form fra STS’en)
     * Sideeffekt: Det eksisterende idkort slettes ikke fra cachen og kan fortsat benyttes til at kalde en test-service (fx NTS) via GW
     */
    @Test
    public void testInvalidRequestWithIdCardInCache() throws Exception  {
        // GIVEN
        String existingIdcardRequest = getResourceString("../sts/servlet/OIOH3BST2SOSI-with-nameid-claim-request.xml");
        Document responseDoc = sendRequest(existingIdcardRequest, CREATE_IDCARD_FROM_BST_URL, true);
        OIOBSTSAMLAssertionToIDCardResponse response = parseResponse(responseDoc);
        assertFalse(response.getFaultString(), response.isFault());
        String invalidRequestBody = getResourceString("../sts/servlet/OIOH3BST2SOSI-request-with-invalid-signature.xml");

        // WHEN
        responseDoc = sendRequest(invalidRequestBody, CREATE_IDCARD_FROM_BST_URL, false);

        // THEN
        response = parseResponse(responseDoc);
        assertTrue(response.isFault());
        assertEquals("Unable to validate envelope signature", response.getFaultString());
        assertNull(response.getIDCard());
        String getValidIdCardRequestBody = getResourceString("getValidIdCard-Mads-Skjern-request.xml");
        sendRequest(getValidIdCardRequestBody, REQUEST_URL, true); // Fails if idcard is not in cache
    }

    /**
     * GW-T6
     * Prækondition: Bruger har ikke et idkort i GW-cachen i forvejen
     * Input: Invalidt createIdCardFromBST request
     * Output: Fejlsvar uden idkort (videresendt i uforandret form fra STS’en)
     * Sideeffekt: Ingen (brugeren har fortsat ikke nogen idkort i cachen)
     */
    @Test
    public void testInvalidRequestWithoutIdCardInCache() throws Exception  {
        // GIVEN
        String invalidRequestBody = getResourceString("../sts/servlet/OIOH3BST2SOSI-request-with-invalid-signature.xml");

        // WHEN
        Document responseDoc = sendRequest(invalidRequestBody, CREATE_IDCARD_FROM_BST_URL, false);

        // THEN
        OIOBSTSAMLAssertionToIDCardResponse response = parseResponse(responseDoc);
        assertTrue(response.isFault());
        assertEquals("Unable to validate envelope signature", response.getFaultString());
        assertNull(response.getIDCard());
        String getValidIdCardRequestBody = getResourceString("getValidIdCard-Korsbaek-request.xml");
        sendRequestAndAssertFault(getValidIdCardRequestBody, REQUEST_URL, "sosigw_no_valid_idcard_in_cache");
    }

    /**
     * GW-T7
     * Prækondition: Bruger har ikke et idkort i GW-cachen i forvejen
     * Input: Validt createIdCardFromBST request med krypteret BST uden angivelse af et ”sosi:SubjectNameID” claim
     * Output: Fejlsvar fra SOSI-GW ("createIdCardFromBST with encrypted bootstraptoken must include a SubjectNameID claim")
     * Sideeffekt: Ingen (brugeren har fortsat ikke nogen idkort i cachen)
     */
    @Test
    public void testValidRequestWithoutIdCardInCache() throws Exception {
        // GIVEN
        String requestBody = getResourceString("../sts/servlet/OIO3BST2SOSI-without-nameid-claim-request.xml");

        // WHEN
        Document responseDoc = sendRequest(requestBody, CREATE_IDCARD_FROM_BST_URL, false);

        // THEN
        assertEquals("Error expected", 400, statusCode);
        Node responseBody = responseDoc.getElementsByTagName("body").item(0);
        assertNotNull(responseBody);
        assertEquals("createIdCardFromBST with encrypted bootstraptoken must include a SubjectNameID claim", responseBody.getTextContent());
    }

    private Document sendRequest(String requestBody, String url, boolean failOnError) throws IOException {
        String responseString = callServer(requestBody, url, failOnError);
        assertNotNull(responseString);
        return XmlUtil.readXml(new Properties(), responseString, false);
    }

    private void sendRequestAndAssertFault(String requestBody, String url, String expectedFaultCode) throws IOException {
        Document response = sendRequest(requestBody, url, false);
        Node faultCodeNode = response.getElementsByTagNameNS("*", "FaultCode").item(0);
        assertNotNull("Expected fault", faultCodeNode);
        assertEquals(expectedFaultCode, faultCodeNode.getTextContent());
    }

    private void validateResponse(Document responseDoc, String cacheKey) throws Exception {
        OIOBSTSAMLAssertionToIDCardResponse response = parseResponse(responseDoc);
        assertFalse(response.getFaultString(), response.isFault());
        assertEquals("Response should not be signed", 0, responseDoc.getDocumentElement().getElementsByTagName(DSTags.SIGNATURE_PREFIXED).getLength());
        assertNotNull(response.getIDCard());
        assertTrue(response.getIDCard() instanceof  UserIDCard);
        UserIDCard idCard = (UserIDCard) response.getIDCard();
        validateIdCard(idCard);
        validateAgainstNTS(idCard, cacheKey);
    }

    private OIOBSTSAMLAssertionToIDCardResponse parseResponse(Document response) {
        return new OIOBSTSAMLAssertionToIDCardResponseModelBuilder().build(response);
    }

    private void validateIdCard(UserIDCard idCard) {
        assertEquals("0707764321", idCard.getUserInfo().getCPR());
        assertEquals("NSTSSenAtre", idCard.getUserInfo().getGivenName());
        assertEquals("Sørensen", idCard.getUserInfo().getSurName());
        assertEquals("7170", idCard.getUserInfo().getRole());
        assertEquals("004NX", idCard.getUserInfo().getAuthorizationCode());
        assertEquals("20301823", idCard.getSystemInfo().getCareProvider().getID());
        assertEquals("Korsbæk Kommune", idCard.getSystemInfo().getCareProvider().getOrgName());
        assertTrue(idCard.getIssuer().contains("STS")); // Afhænger af test miljø, men indeholder altid STS.
    }

    private void validateAgainstNTS(UserIDCard idCard, String cacheKey) {
        AssertionType assertionType = IDCardToSamlAssertion.build(idCard);
        assertionType.getSubject().getNameID().setFormat(SubjectIdentifierTypeValues.OTHER);
        assertionType.getSubject().getNameID().setValue(cacheKey);
        Security security = new Security();
        security.setTimestamp(IdCardTestUtils.makeTimestamp());
        security.setAssertion(assertionType);
        try {
            String ntsResult = nts.verify(security, medcomHeader, "payload");
            assertNotNull(ntsResult);
        } catch (dk.nspop.nts._2013._05.DGWSFault dgwsFault) {
            Assert.fail(dgwsFault.getFaultInfo());
        }
    }

    String getResourceString(String resource) {
        return new Scanner(this.getClass().getResourceAsStream(resource), "UTF-8").useDelimiter("\\A").next();
    }
}
