package com.trifork.sosigw.inttest;

import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.HttpURLConnection;
import java.net.URISyntaxException;
import java.net.URL;
import java.security.InvalidKeyException;
import java.security.KeyStore;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.NoSuchProviderException;
import java.security.PrivateKey;
import java.security.Signature;
import java.security.SignatureException;
import java.security.UnrecoverableKeyException;
import java.security.cert.CertificateException;
import java.util.Base64;
import java.util.Properties;
import java.util.UUID;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.commons.httpclient.HttpClient;
import org.apache.commons.httpclient.methods.PostMethod;
import org.apache.commons.httpclient.methods.StringRequestEntity;
import org.apache.commons.io.IOUtils;
import org.apache.log4j.Logger;
import org.junit.Before;
import org.junit.Test;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

import com.trifork.sosigw.util.Util;
import com.trifork.sosigw.ws.IdCardTestUtils;

import dk.sosi.seal.SOSIFactory;
import dk.sosi.seal.model.AuthenticationLevel;
import dk.sosi.seal.model.CareProvider;
import dk.sosi.seal.model.IDCard;
import dk.sosi.seal.model.Request;
import dk.sosi.seal.model.SecurityTokenRequest;
import dk.sosi.seal.model.SignatureUtil;
import dk.sosi.seal.model.UserIDCard;
import dk.sosi.seal.model.UserInfo;
import dk.sosi.seal.model.constants.SubjectIdentifierTypeValues;
import dk.sosi.seal.pki.SOSIFederation;
import dk.sosi.seal.pki.impl.HashMapCertificateCache;
import dk.sosi.seal.vault.CredentialVaultException;
import dk.sosi.seal.vault.FileBasedCredentialVault;
import dk.sosi.seal.xml.XmlUtil;

/**
 * This test class is dependent on a niab running with sosigw.mode=test, also sosigw console must have been configured
 * according to the "Golden image" documentation.
 */
public class ProxyIdCardIT extends AbstractIT {
    private static final Logger logger = Logger.getLogger(ProxyIdCardIT.class);
    private static final String HTTP_STS_SECURITY_TOKEN_SERVICE = "http://test1-cnsp.ekstern-test.nspop.dk:8080/sts/services/SecurityTokenService";
    private String signonUrl = System.getProperty("signonUrl", "http://localhost:8080/sosigw/service/sosigw");
    private String proxyUrl = System.getProperty("proxyUrl", "http://localhost:8080/sosigw/proxy/soap-request");

    private static final String MOCES_VAULT = "/etc/Sonja_Bach_Laege.jks";
    private static final String MOCES_VAULT_PASSWORD = "Test1234";
    private static final String MOCES_CERTIFICATE_ALIAS = "SOSI:ALIAS_SYSTEM";

    @Test
    public void getMedicinCardWithSignedRequestThroughSosiProxy() throws Exception {
        String fullIdCardSoapResponse = createLevel4IdCard();
        String idCardHeader = fullIdCardSoapResponse.substring(0, fullIdCardSoapResponse.length() - 51); // remove
        // "</soapenv:Header><soapenv:Body/></soapenv:Envelope>"

        // add SOAP action
        idCardHeader = idCardHeader.replace("<soapenv:Header>",
                "<soapenv:Header><Action xmlns=\"http://schemas.xmlsoap.org/ws/2004/08/addressing\">http://www.dkma.dk/medicinecard/xml.schema/2015/01/01/E1#GetMedicineCard</Action>");

        final String signedIdCardRequest = idCardHeader + loadFileContent("getMedicineCard_request.xml");
        System.out.println("\n----request:\n" + signedIdCardRequest + "\n\n");

        final String requestIdCardResp = callServer(signedIdCardRequest, proxyUrl);

        System.out.println("Server response to proxy request: \n" + requestIdCardResp);
    }

    @Test
    public void getMedicinCardWithUnSignedRequestThroughSosiProxy() throws Exception {
        logger.info("Creating level1 idcard");
        String idCardAssertion = createLevel1IdCard();

        KeyStore keystore = getMocesKeyStore();
        String idCardSignatureValue = getSignatureValue(keystore, idCardAssertion);
        String encodedCertificate = XmlUtil.toBase64(keystore.getCertificate("sosi:alias_system").getEncoded());

        signIdCardAndInsertCardInGatewayCache(idCardAssertion, idCardSignatureValue, encodedCertificate);

        logger.info("Preparing gatewayProxyMedecineCardRequest from callGatewayProxyMedecineCardRequest-template.xml");
        final String callGatewayProxyMedecineCardRequestTemplate = loadFileContent(
                "callGatewayProxyMedecineCardRequest-template.xml");
        String callGatewayProxyMedecineCardRequest = callGatewayProxyMedecineCardRequestTemplate
                .replace("#IDCARD_ASSERTION#", idCardAssertion).replace("#MESSAGE_ID#", UUID.randomUUID().toString());

        logger.info("Calling server " + proxyUrl + " with XML\n" + callGatewayProxyMedecineCardRequest);
        String callGatewayProxyMedecineCardResponse = callServer(callGatewayProxyMedecineCardRequest, proxyUrl,
                new RequestHeader("SOAPAction", "\"http://www.dkma.dk/medicinecard/xml.schema/2015/01/01/E1#GetMedicineCard\""));
        logger.debug("callGatewayProxyMedecineCardResponse: \n" + callGatewayProxyMedecineCardResponse + "\n");

        System.out.println("Server response to proxy request: \n" + callGatewayProxyMedecineCardResponse);
    }

    private String loadFileContent(String filename) throws IOException {
        return Util.readAllFromStreamAsString(this.getClass().getResourceAsStream(filename));
    }

    private void signIdCardAndInsertCardInGatewayCache(String idCardAssertion, String idCardSignatureValue, String encodedCertificate)
            throws IOException {
        logger.debug("Preparing signIdCardRequest from callGatewaySignIdCard-request-template.xml");
        String signIdCardRequestTemplate = loadFileContent("callGatewaySignIdCard-request-template.xml"); // this.getClass().getResourceAsStream("callGatewaySignIdCard-request-template.xml"));
        String signIdCardRequest = signIdCardRequestTemplate
                .replace("#SIGNATURE_VALUE#", idCardSignatureValue)
                .replace("#IDCARD_ASSERTION#", idCardAssertion)
                .replace("#ENCODED_CERTIFICATE#", encodedCertificate);
        logger.debug("Calling server " + signonUrl + " with XML\n" + signIdCardRequest);

        String signIdCardResponse = callServer(signIdCardRequest, signonUrl); // this will put the idcard in the
        // gateway cache
        logger.debug("signIdCardResponse: \n" + signIdCardResponse + "\n");
    }

    private String getSignatureValue(KeyStore keystore, String idCardAssertion) throws Exception {
        logger.debug("Preparing requestSignatureXml from callGatewayRequestSignature-template.xml");
        String requestSignatureXmlTemplate = loadFileContent("callGatewayRequestSignature-template.xml");
        String requestSignatureXml = requestSignatureXmlTemplate.replaceAll("#IDCARD_ASSERTION#", idCardAssertion);
        logger.debug("Calling server " + signonUrl + " with XML\n" + requestSignatureXml);

        String requestIdCardResp = callServer(requestSignatureXml, signonUrl);

        logger.debug("Parsing DigestValue from response");
        Matcher matcher = Pattern.compile("<(.*:)?DigestValue>([^<]*)</(.*:)?DigestValue>").matcher(requestIdCardResp);
        assertTrue(requestIdCardResp + " contains DigestValue", matcher.find());
        final String digestValue = matcher.group(2).trim();
        return sign(keystore, digestValue);
    }

    private String sign(KeyStore keystore, String digestValue) throws NoSuchAlgorithmException, KeyStoreException, IOException,
            UnrecoverableKeyException, InvalidKeyException, SignatureException {
        PrivateKey key = (PrivateKey)keystore.getKey(MOCES_CERTIFICATE_ALIAS, MOCES_VAULT_PASSWORD.toCharArray());

        byte[] signatureValue = IdCardTestUtils.signDigest(Base64.getDecoder().decode(digestValue), key);

        return Base64.getEncoder().encodeToString(signatureValue);
    }

    private KeyStore getMocesKeyStore() throws KeyStoreException, IOException, NoSuchAlgorithmException, CertificateException {
        KeyStore keystore = KeyStore.getInstance("JKS");
        keystore.load(getClass().getResourceAsStream(MOCES_VAULT), MOCES_VAULT_PASSWORD.toCharArray());
        return keystore;
    }

    private String createLevel1IdCard() throws IOException, CredentialVaultException, URISyntaxException {

        Properties props = SignatureUtil.setupCryptoProviderForJVM();
        props.setProperty(SOSIFactory.PROPERTYNAME_SOSI_VALIDATE_ENHANCED, "false");
        props.setProperty(SOSIFactory.PROPERTYNAME_SOSI_VALIDATE, "false");

        SOSIFederation federation = new SOSIFederation(props, new HashMapCertificateCache());

        FileBasedCredentialVault vault = new FileBasedCredentialVault(props, getMocesVaultFile(), "Test1234");

        SOSIFactory factory = new SOSIFactory(federation, vault, props);

        SecurityTokenRequest stReq = factory.createNewSecurityTokenRequest();

        String cpr = "0309691444";
        CareProvider careProvider = new CareProvider(SubjectIdentifierTypeValues.CVR_NUMBER, "46837428", "test");
        UserInfo userInfo = new UserInfo(cpr, "Sonja", "Bach", "test@test.com", "developer", "Doctor", "NS363");
        UserIDCard card = factory.createNewUserIDCard("FMK-testclient", userInfo, careProvider,
                AuthenticationLevel.NO_AUTHENTICATION, null, null, vault.getSystemCredentialPair().getCertificate(),
                "Test");
        stReq.setIDCard(card);
        Element idCardNode = stReq.getIDCard().serialize2DOMDocument(factory, stReq.serialize2DOMDocument());
        String xml = XmlUtil.node2String(idCardNode, false, false);

        return xml;
    }

    private File getMocesVaultFile() throws URISyntaxException {
        return new File(getClass().getResource(MOCES_VAULT).toURI());
    }

    private String createLevel4IdCard() throws NoSuchAlgorithmException, NoSuchProviderException, InvalidKeyException,
            SignatureException, IOException, CredentialVaultException, URISyntaxException {
        String cpr = "0309691444";
        Properties props = SignatureUtil.setupCryptoProviderForJVM();
        props.setProperty(SOSIFactory.PROPERTYNAME_SOSI_VALIDATE_ENHANCED, "false");
        props.setProperty(SOSIFactory.PROPERTYNAME_SOSI_VALIDATE, "false");

        SOSIFederation federation = new SOSIFederation(props, new HashMapCertificateCache());
        FileBasedCredentialVault vault = new FileBasedCredentialVault(props, getMocesVaultFile(), "Test1234");
        SOSIFactory factory = new SOSIFactory(federation, vault, props);

        SecurityTokenRequest stReq = factory.createNewSecurityTokenRequest();

        UserInfo userInfo = new UserInfo(cpr, "Sonja", "Bach", "test@test.com", "developer", "Doctor", "NS363");
        CareProvider careProvider = new CareProvider(SubjectIdentifierTypeValues.CVR_NUMBER, "46837428", "test");
        UserIDCard card = factory.createNewUserIDCard("SOSITEST", userInfo, careProvider,
                AuthenticationLevel.MOCES_TRUSTED_USER, null, null, vault.getSystemCredentialPair().getCertificate(),
                "Test");

        stReq.setIDCard(card);
        byte[] bytesForSigning = stReq.getIDCard().getBytesForSigning(stReq.serialize2DOMDocument());

        Signature jceSign = Signature.getInstance("SHA1withRSA",
                SignatureUtil.getCryptoProvider(props, SOSIFactory.PROPERTYNAME_SOSI_CRYPTOPROVIDER_SHA1WITHRSA));
        PrivateKey key = vault.getSystemCredentialPair().getPrivateKey();
        jceSign.initSign(key);
        jceSign.update(bytesForSigning);
        String signature = XmlUtil.toBase64(jceSign.sign());

        stReq.getIDCard().injectSignature(signature, vault.getSystemCredentialPair().getCertificate());

        String xml = XmlUtil.node2String(stReq.serialize2DOMDocument(), false, true);
        String stsUrl = HTTP_STS_SECURITY_TOKEN_SERVICE;
        String stResXml = sendRequest(stsUrl, "", xml, false);

        SecurityTokenRequest stRes = factory.deserializeSecurityTokenRequest(stResXml);
        IDCard idCard = stRes.getIDCard();

        Request sreq = factory.createNewRequest(false, "flowId");
        sreq.setIDCard(idCard);
        Document reqdoc = sreq.serialize2DOMDocument();

        return XmlUtil.node2String(reqdoc, false, true);
    }

    public static String sendRequest(String url, String action, String docXml, boolean failOnError) throws IOException {
        URL u = new URL(url);
        HttpURLConnection uc = (HttpURLConnection)u.openConnection();
        uc.setDoOutput(true);
        uc.setDoInput(true);
        uc.setRequestMethod("POST");
        uc.setRequestProperty("SOAPAction", "\"" + action + "\"");
        uc.setRequestProperty("Content-Type", "text/xml; encoding=utf-8");
        OutputStream os = uc.getOutputStream();

        IOUtils.write(docXml, os);
        os.flush();

        InputStream is;
        if (uc.getResponseCode() != 200) {
            is = uc.getErrorStream();
        } else {
            is = uc.getInputStream();
        }
        String res = IOUtils.toString(is);

        is.close();
        if (uc.getResponseCode() != 200 && (uc.getResponseCode() != 500 || failOnError)) {
            throw new IOException(res);
        }
        os.close();
        uc.disconnect();

        return res;
    }
}
