/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 * $HeadURL$
 * $Id$
 */
package com.trifork.sosigw.sts.servlet;

import com.trifork.sosigw.Factory;
import com.trifork.sosigw.api.IDCardCache;
import com.trifork.sosigw.api.IDCardState;
import com.trifork.sosigw.cluster.IDCardClusterCacheEntry;
import com.trifork.sosigw.cluster.IDCardClusterCacheImpl;
import com.trifork.sosigw.config.StaticConfigImpl;
import com.trifork.sosigw.sts.STSProxy;
import com.trifork.sosigw.util.Util;
import dk.sdsd.nsp.slalog.impl.SLALoggerImpl;
import dk.sosi.gw._2007_09.DGWSFault;
import dk.sosi.seal.model.AuthenticationLevel;
import dk.sosi.seal.model.CareProvider;
import dk.sosi.seal.model.OIOBSTSAMLAssertionToIDCardResponse;
import dk.sosi.seal.model.SystemInfo;
import dk.sosi.seal.model.UserIDCard;
import dk.sosi.seal.model.UserInfo;
import dk.sosi.seal.model.constants.DSTags;
import dk.sosi.seal.modelbuilders.OIOBSTSAMLAssertionToIDCardResponseModelBuilder;
import dk.sosi.seal.xml.XmlUtil;
import org.junit.AfterClass;
import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Mockito;
import org.w3c.dom.Document;

import javax.servlet.ServletException;
import javax.servlet.ServletInputStream;
import javax.servlet.ServletOutputStream;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.nio.charset.StandardCharsets;
import java.util.Properties;
import java.util.Scanner;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;
import static org.junit.Assert.assertTrue;
import static org.mockito.Matchers.any;
import static org.mockito.Matchers.anyInt;
import static org.mockito.Mockito.*;


public class CreateIdCardFromBSTTest {
    private static STSProxy originalStsService;
    private CreateIdCardFromBSTServlet createIdCardFromBSTServlet;
    private HttpServletRequest httpRequest;
    private HttpServletResponse httpResponse;
    private InputStream httpRequestInputStream;
    private ByteArrayOutputStream responseOutputStream;

    @BeforeClass
    public static void saveOriginalStsService() {
        StaticConfigImpl.initSettings();
        Util.setSLALogger(new SLALoggerImpl());
        originalStsService = CreateIdCardFromBSTRequestHandler.stsService;
    }

    @AfterClass
    public static void revertStsService() {
        CreateIdCardFromBSTRequestHandler.stsService = originalStsService;
    }

    @Before
    public void setup() throws IOException {
        Factory.setTheIDCardCache(new IDCardClusterCacheImpl(null));
        createIdCardFromBSTServlet = new CreateIdCardFromBSTServlet();
        httpRequest = Mockito.mock(HttpServletRequest.class);
        httpRequestInputStream = new ByteArrayInputStream("".getBytes(StandardCharsets.UTF_8));
        when(httpRequest.getInputStream()).thenReturn(new ServletInputStream() {
            @Override
            public int read() throws IOException {
                return httpRequestInputStream.read();
            }
        });

        httpResponse = Mockito.mock(HttpServletResponse.class);
        responseOutputStream = new ByteArrayOutputStream();
        when(httpResponse.getOutputStream()).thenReturn(new ServletOutputStream() {
            @Override
            public void write(int b) {
                responseOutputStream.write(b);
            }
        });
    }

    /**
     * GW-T1
     * Prækondition: Bruger har ikke et idkort i GW-cachen i forvejen
     * Input: Validt createIdCardFromBST request uden angivelse af et ”sosi:SubjectNameID” claim
     * Output: Validt svar med det nyudstedte idkort i usigneret form
     * Sideeffekt: Det nyudstedte idkort er gemt i cachen under SubjectNameID fra bootstraptokenet og kan benyttes til at kalde en test-service (fx NTS) via GW
     */
    @Test
    public void testValidRequestWithoutNameIdClaimEmptyCache() throws Exception  {
        setupStsSuccessResponse();
        final IDCardCache idCardCache = Factory.getIDCardCache();
        assertEquals(0, idCardCache.size());

        Document responseDoc = sendRequest("OIOH3BST2SOSI-without-nameid-claim-request.xml");

        validateResponse(idCardCache, responseDoc, "KorsbaekKommune\\MSK");
    }

    /**
     * GW-T2
     * Prækondition: Bruger har ikke et idkort i GW-cachen i forvejen
     * Input: Validt createIdCardFromBST request med angivelse af et ”sosi:SubjectNameID” claim (som er forskelligt fra SubjectNameID i BST)
     * Output: Validt svar med det nyudstedte idkort i usigneret form
     * Sideeffekt: Det nyudstedte idkort er gemt i cachen under SubjectNameID fra claimet og kan benyttes til at kalde en test-service (fx NTS) via GW
     */
    @Test
    public void testValidRequestWithNameIdClaimEmptyCache() throws Exception  {
        setupStsSuccessResponse();
        final IDCardCache idCardCache = Factory.getIDCardCache();
        assertEquals(0, idCardCache.size());

        Document responseDoc = sendRequest("OIOH3BST2SOSI-with-nameid-claim-request.xml");

        validateResponse(idCardCache, responseDoc, "Mads_Skjern");
    }

    /**
     * GW-T3
     * Prækondition: Bruger har et idkort i GW-cachen i forvejen
     * Input: Validt createIdCardFromBST request uden angivelse af et ”sosi:SubjectNameID” claim
     * Output: Validt svar med det nyudstedte idkort i usigneret form
     * Sideeffekt: Det nyudstedte idkort er gemt i cachen under SubjectNameID fra bootstraptokenet, har erstattet det eksisterende idkort og kan benyttes til at kalde en test-service (fx NTS) via GW
     */
    @Test
    public void testValidRequestWithoutNameIdClaimWithIdCardInCache() throws Exception  {
        setupStsSuccessResponse();
        final IDCardCache idCardCache = Factory.getIDCardCache();
        saveTestEntryInCache(idCardCache, "KorsbaekKommune\\MSK");
        assertEquals(1, idCardCache.size());

        Document responseDoc = sendRequest("OIOH3BST2SOSI-without-nameid-claim-request.xml");

        validateResponse(idCardCache, responseDoc, "KorsbaekKommune\\MSK");
    }

    /**
     * GW-T4
     * Prækondition: Bruger har et idkort i GW-cachen i forvejen
     * Input: Validt createIdCardFromBST request med angivelse af et ”sosi:SubjectNameID” claim (som er forskelligt fra SubjectNameID i BST)
     * Output: Validt svar med det nyudstedte idkort i usigneret form
     * Sideeffekt: Det nyudstedte idkort er gemt i cachen under SubjectNameID fra claimet, har erstattet det eksisterende idkort og kan benyttes til at kalde en test-service (fx NTS) via GW
     */
    @Test
    public void testValidRequestWithNameIdClaimWithIdCardInCache() throws Exception  {
        setupStsSuccessResponse();
        final IDCardCache idCardCache = Factory.getIDCardCache();
        saveTestEntryInCache(idCardCache, "Mads_Skjern");
        assertEquals(1, idCardCache.size());

        Document responseDoc = sendRequest("OIOH3BST2SOSI-with-nameid-claim-request.xml");

        validateResponse(idCardCache, responseDoc, "Mads_Skjern");
    }

    /**
     * GW-T5
     * Prækondition: Bruger har et idkort i GW-cachen i forvejen
     * Input: Invalidt createIdCardFromBST request
     * Output: Fejlsvar uden idkort (videresendt i uforandret form fra STS’en)
     * Sideeffekt: Det eksisterende idkort slettes ikke fra cachen og kan fortsat benyttes til at kalde en test-service (fx NTS) via GW
     */
    @Test
    public void testInvalidRequestWithIdCardInCache() throws Exception  {
        setupStsErrorResponse();
        final IDCardCache idCardCache = Factory.getIDCardCache();
        IDCardClusterCacheEntry testCacheEntry = saveTestEntryInCache(idCardCache, "KorsbaekKommune\\MSK");
        assertEquals(1, idCardCache.size());

        Document responseDoc = sendRequest("OIOH3BST2SOSI-request-with-invalid-signature.xml");

        OIOBSTSAMLAssertionToIDCardResponse response = parseResponse(responseDoc);
        assertTrue(response.isFault());
        assertNull(response.getIDCard());
        validateCache(idCardCache, "KorsbaekKommune\\MSK", testCacheEntry.getIdCard(), testCacheEntry.getState());
    }

    /**
     * GW-T6
     * Prækondition: Bruger har ikke et idkort i GW-cachen i forvejen
     * Input: Invalidt createIdCardFromBST request
     * Output: Fejlsvar uden idkort (videresendt i uforandret form fra STS’en)
     * Sideeffekt: Ingen (brugeren har fortsat ikke nogen idkort i cachen)
     */
    @Test
    public void testInvalidRequestWithoutIdCardInCache() throws Exception  {
        setupStsErrorResponse();
        final IDCardCache idCardCache = Factory.getIDCardCache();
        assertEquals(0, idCardCache.size());

        Document responseDoc = sendRequest("OIOH3BST2SOSI-request-with-invalid-signature.xml");

        OIOBSTSAMLAssertionToIDCardResponse response = parseResponse(responseDoc);
        assertTrue(response.isFault());
        assertNull(response.getIDCard());
        assertEquals(0, idCardCache.size());
    }

    @Test
    public void testValidEncryptedRequestWithNameIdClaim() throws Exception {
        setupStsSuccessResponse();
        final IDCardCache idCardCache = Factory.getIDCardCache();
        assertEquals(0, idCardCache.size());

        Document responseDoc = sendRequest("OIO3BST2SOSI-with-nameid-claim-request.xml");

        validateResponse(idCardCache, responseDoc, "Mads_Skjern");
    }

    @Test
    public void testValidEncryptedRequestWithoutNameIdClaim() throws ServletException, IOException {
        ArgumentCaptor<String> errorMessageCaptor = ArgumentCaptor.forClass(String.class);
        doNothing().when(httpResponse).sendError(anyInt(), errorMessageCaptor.capture());

        setupStsSuccessResponse();
        final IDCardCache idCardCache = Factory.getIDCardCache();
        assertEquals(0, idCardCache.size());


        postRequest("OIO3BST2SOSI-without-nameid-claim-request.xml");
        assertEquals("createIdCardFromBST with encrypted bootstraptoken must include a SubjectNameID claim", errorMessageCaptor.getValue());
        assertEquals(0, idCardCache.size());
    }

    private void setupStsSuccessResponse() {
        STSProxy stsProxy = mock(STSProxy.class);
        CreateIdCardFromBSTRequestHandler.stsService = stsProxy;
        String stsResponse = getResourceString("XXXBST2SOSI-response.xml");
        when(stsProxy.callBST2SOSI(any())).thenReturn(stsResponse);
    }

    private void setupStsErrorResponse() {
        STSProxy stsProxy = mock(STSProxy.class);
        CreateIdCardFromBSTRequestHandler.stsService = stsProxy;
        String stsResponse = getResourceString("XXXBST2SOSI-error-response.xml");
        when(stsProxy.callBST2SOSI(any())).thenReturn(stsResponse);
    }

    private IDCardClusterCacheEntry saveTestEntryInCache(IDCardCache idCardCache, String cacheKey) {
        IDCardClusterCacheEntry cachedIDCardEntry = idCardCache.ensureIDCard(cacheKey);
        final SystemInfo sysInfo = new SystemInfo(new CareProvider("medcom:skscode", "test", "test"), "test");
        final UserInfo userInfo = new UserInfo("1111111118", "Darkwing", "Duck", null, null, "role", null);
        UserIDCard userIDCard = new UserIDCard("1.0.1", AuthenticationLevel.MOCES_TRUSTED_USER, "test", sysInfo, userInfo, "e", "2", null, null);
        cachedIDCardEntry.setIdCard(userIDCard);
        idCardCache.updateEntryState(cachedIDCardEntry, IDCardState.SIGNED);
        idCardCache.entryUpdatedLocally(cachedIDCardEntry);
        return cachedIDCardEntry;
    }

    private void validateResponse(IDCardCache idCardCache, Document responseDoc, String cacheKey) throws Exception {
        OIOBSTSAMLAssertionToIDCardResponse response = parseResponse(responseDoc);
        assertFalse(response.isFault());
        assertEquals("Response should not be signed", 0, responseDoc.getDocumentElement().getElementsByTagName(DSTags.SIGNATURE_PREFIXED).getLength());
        assertNotNull(response.getIDCard());
        assertTrue(response.getIDCard() instanceof  UserIDCard);
        UserIDCard idCard = (UserIDCard) response.getIDCard();
        validateIdCard(idCard);
        validateCache(idCardCache, cacheKey, idCard, IDCardState.SIGNED);
    }

    private void validateIdCard(UserIDCard idCard) {
        // From XXXBST2SOSI-response.xml
        assertEquals("0707764321", idCard.getUserInfo().getCPR());
        assertEquals("NSTSSenAtre", idCard.getUserInfo().getGivenName());
        assertEquals("Sorensen", idCard.getUserInfo().getSurName());
        assertEquals("7170", idCard.getUserInfo().getRole());
        assertEquals("004NX", idCard.getUserInfo().getAuthorizationCode());
        assertEquals("20301823", idCard.getSystemInfo().getCareProvider().getID());
        assertEquals("Korsbæk Kommune", idCard.getSystemInfo().getCareProvider().getOrgName());
        assertEquals("TESTSTS", idCard.getIssuer());
        assertEquals("Mads_Skjern", idCard.getAlternativeIdentifier());
    }

    private void validateCache(IDCardCache idCardCache, String cacheKey, UserIDCard expectedIdCard, IDCardState expectedState) {
        assertEquals(1, idCardCache.size());
        IDCardClusterCacheEntry cacheEntry = idCardCache.findIDCardOrNull(cacheKey);
        assertNotNull(cacheEntry);
        assertEquals(expectedState, cacheEntry.getState());
        assertEquals(expectedIdCard, cacheEntry.getIdCard());
    }

    private OIOBSTSAMLAssertionToIDCardResponse parseResponse(Document response) {
        return new OIOBSTSAMLAssertionToIDCardResponseModelBuilder().build(response);
    }

    private Document sendRequest(String requestResource) throws ServletException, IOException {
        postRequest(requestResource);
        String responseString = responseOutputStream.toString();
        return XmlUtil.readXml(new Properties(), responseString, false);
    }

    private void postRequest(String requestResource) throws ServletException, IOException {
        httpRequestInputStream = CreateIdCardFromBSTTest.class.getResourceAsStream(requestResource);
        createIdCardFromBSTServlet.doPost(httpRequest, httpResponse);
    }

    String getResourceString(String resource) {
        return new Scanner(CreateIdCardFromBSTTest.class.getResourceAsStream(resource), "UTF-8").useDelimiter("\\A").next();
    }
}
